"""
The package ``syn`` provides components necessary to generate synthetic data.

===============================================================================
Overview
===============================================================================
The package ``syn`` (short for **syn**\ thetic) contains components necessary
for generating synthetic data. This is primarily used for testing the
algorithm. However, it could be used for benchmarking algorithms or in other
cases.
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Oct 15, 2014 18:16:33 EDT$"

__all__ = ["data"]


from nanshe.syn import data
